// Author: Professor William Killian
// Date: 2019 July
// Class: CSCI 362.93 - Data Structures
// Assignment: 8 - Graph Algorithms
//
// Description:
//
// A basic Disjoint Set class

#include "DisjointSet.hpp"

#include <numeric>
#include <utility>

DisjointSet::DisjointSet (size_t N) noexcept
  : m_parent (N, 0), m_rank (N, 0), m_sizes (N, 1), m_setcount (N)
{
  std::iota (m_parent.begin (), m_parent.end (), 0);
}

size_t
DisjointSet::findSet (size_t i) noexcept
{
  if (m_parent[i] == i)
  {
    return i;
  }
  /* original:
  return findSet (m_parent[i]);
  */

  // optimization (compression):
  return (m_parent[i] = findSet (m_parent[i]));
}

bool
DisjointSet::isSameSet (size_t i, size_t j) noexcept
{
  return findSet (i) == findSet (j);
}

void
DisjointSet::unionSet (size_t i, size_t j) noexcept
{
  if (isSameSet (i, j))
  {
    return;
  }
  // find both representative items
  size_t x = findSet (i);
  size_t y = findSet (j);

  // optimization (rank): keep x 'shorter' than y
  if (m_rank[x] > m_rank[y])
  {
    std::swap (x, y);
  }
  // set x under y
  m_parent[x] = y;

  // optimization (rank):
  if (m_rank[x] == m_rank[y])
  {
    ++m_rank[y];
  }

  // combine set sizes at y
  m_sizes[y] += m_sizes[x];

  // a union reduces numSets
  --m_setcount;
}

size_t
DisjointSet::size () const noexcept
{
  return m_parent.size ();
}

size_t
DisjointSet::size (size_t i) noexcept
{
  return m_sizes[findSet (i)];
}

size_t
DisjointSet::set_count () const noexcept
{
  return m_setcount;
}
