// Author: Professor William Killian
// Date: 2019 July
// Class: CSCI 362.93 - Data Structures
// Assignment: 8 - Graph Algorithms
//
// Description:
//
// A basic graph class which uses adjacency lists

#include "Graph.hpp"

#include <algorithm>

bool
operator< (Edge const& e1, Edge const& e2)
{
  return e1.weight < e2.weight;
}

bool
operator== (Edge const& e1, Edge const& e2)
{
  return e1.src == e2.src && e1.dest == e2.dest;
}

bool
operator!= (Edge const& e1, Edge const& e2)
{
  return !(e1 == e2);
}

namespace std
{

size_t
hash<Edge>::operator() (Edge const& e) const
{
  return (e.src << ((sizeof (e.dest) >> 1) << 3)) | e.dest;
}
}

void
Graph::addEdge (size_t src, size_t dest, double weight)
{
  mVertices.insert (src);
  if (mEdges[dest].empty ())
  {
    mEdges[dest].reserve (10);
  }
  mVertices.insert (dest);
  mEdges[src].emplace_back (src, dest, weight);
}

Graph::EdgeList
Graph::edges () const
{
  std::vector<Edge> all;
  for (auto&& [src, es] : mEdges)
  {
    all.insert (all.end (), es.begin (), es.end ());
  }
  return all;
}

Graph::EdgeList const&
Graph::edges (size_t src) const
{
  return mEdges.at (src);
}

Graph::VertexList
Graph::vertices () const
{
  VertexList all (mVertices.begin (), mVertices.end ());
  std::sort (all.begin (), all.end ());
  return all;
}

size_t
Graph::size () const
{
  return mVertices.size ();
}
