// Author: Professor William Killian
// Date: 2019 July
// Class: CSCI 362.93 - Data Structures
// Assignment: 8 - Graph Algorithms
//
// Description:
//
// A basic graph class which uses adjacency lists

#ifndef GRAPH_HPP_
#define GRAPH_HPP_

#include <unordered_map>
#include <unordered_set>
#include <vector>

struct Edge
{
  size_t src;
  size_t dest;
  double weight;

  inline Edge (size_t src, size_t dest, double weight)
    : src(src), dest(dest), weight(weight)
  {
  }
};

bool
operator< (Edge const& e1, Edge const& e2);

bool
operator== (Edge const& e1, Edge const& e2);

bool
operator!= (Edge const& e1, Edge const& e2);

namespace std
{
template<>
struct hash<Edge>
{
  size_t
  operator() (Edge const& e) const;
};
}

// Graph class
struct Graph
{
  using EdgeList = std::vector<Edge>;
  using VertexList = std::vector<size_t>;

  // Adds an edge to the graph (u, v) = w
  void
  addEdge (size_t src, size_t dest, double weight);

  // Returns an aggregated list of all edges
  // - Note: has complexity O(N); if you want to make repeated
  //   iterations, assign the returned EdgeList to a variable
  EdgeList
  edges () const;

  // Returns the the edge list for a given vertex
  EdgeList const&
  edges (size_t src) const;

  // Returns a list of vertices within the graph
  VertexList
  vertices () const;

  // Returns the number of vertices
  size_t
  size () const;

private:
  std::unordered_map<size_t, EdgeList> mEdges;
  std::unordered_set<size_t> mVertices;
};

#endif
