// Author: Professor William Killian
// Date: 2019 July
// Class: CSCI 362.93 - Data Structures
// Assignment: 8 - Graph Algorithms
//
// Description:
//
// A min heap class which supports decrease_key, extract_min, and insert operations

#ifndef MIN_HEAP_HPP_
#define MIN_HEAP_HPP_

#include <limits>
#include <unordered_map>
#include <vector>

#include "Graph.hpp"

struct DistanceVertexPair
{
  double dist;
  size_t vertex;

  inline DistanceVertexPair (double dist, size_t vertex) : dist (dist), vertex (vertex)
  {
  }

  inline DistanceVertexPair (size_t vertex)
    : DistanceVertexPair (std::numeric_limits<double>::max (), vertex)
  {
  }

  friend inline bool
  operator< (DistanceVertexPair const& p1, DistanceVertexPair const& p2)
  {
    return p1.dist < p2.dist;
  }

  friend inline bool
  operator== (DistanceVertexPair const& p1, DistanceVertexPair const& p2)
  {
    return p1.vertex == p2.vertex;
  }

};

namespace std
{
template<>
struct hash<DistanceVertexPair>
{
  inline size_t
  operator() (DistanceVertexPair const& p) const
  {
    return p.vertex;
  }
};
}

template<typename T>
class MinHeap
{
  size_t
  lchild (size_t index) const
  {
    return 2 * index + 1;
  }

  size_t
  rchild (size_t index) const
  {
    return 2 * index + 2;
  }

  size_t
  parent (size_t index) const
  {
    return index / 2;
  }

  size_t
  last_interior () const
  {
    return parent (mData.size () - 1);
  }

  size_t
  min_child (size_t index) const
  {
    size_t r = rchild (index);
    size_t l = lchild (index);
    if (r >= mData.size () || mData[l] < mData[r])
    {
      return l;
    }
    else
    {
      return r;
    }
  }

public:

  T
  extract_min ()
  {
    size_t last = last_interior();

    T const min = mData.front ();
    swap (mData.front (), mData.back ());
    mIndices.erase (min);
    mData.pop_back ();

    if (mData.size () <= 1)
    {
      return min;
    }
    for (size_t i = 0; i < last; )
    {
      size_t c = min_child (i);
      if (mData[c] < mData[i])
      {
        swap (mData[c], mData[i]);
        i = c;
      }
      else
      {
        break;
      }
    }
    return min;
  }

  bool
  decrease_key (T const& p)
  {
    size_t i = mIndices[p];
    if (p < mData[i])
    {
      mData[i] = p;
    }
    else
    {
      return false;
    }
    while (i > 0)
    {
      size_t p = parent (i);
      if (mData[i] < mData[p])
      {
        swap (mData[i], mData[p]);
        i = p;
      }
      else
      {
        break;
      }
    }
    return true;
  }

  void
  insert (T const& p)
  {
    mIndices[p] = mData.size ();
    mData.push_back (p);
    decrease_key (p);
  }

  T
  get (T const& p)
  {
    return mData[mIndices[p]];
  }

  bool
  empty () const
  {
    return mData.empty ();
  }

private:
  void
  swap (T& p1, T& p2)
  {
    std::swap (mIndices[p1], mIndices[p2]);
    std::swap (p1, p2);
  }

  std::vector<T> mData;
  std::unordered_map<T, size_t> mIndices;
};

#endif
